from typing import Any, Dict, List, Literal, Optional

from pydantic import BaseModel, Field

from ..._resource import AsyncAPIResource, SyncAPIResource
from ...types.standards import PreparedRequest
from .completions import AsyncCompletions, Completions
from .responses import AsyncResponses, Responses
from ...types.consensus import ReconciliationResponse

class BaseConsensusMixin:

    def _prepare_reconcile(
        self,
        list_dicts: List[Dict[str, Any]],
        reference_schema: Optional[Dict[str, Any]] = None,
        mode: Literal["direct", "aligned"] = "direct",
        idempotency_key: str | None = None,
    ) -> PreparedRequest:
        data = {
            "list_dicts": list_dicts,
            "mode": mode,
        }
        
        if reference_schema is not None:
            data["reference_schema"] = reference_schema
            
        return PreparedRequest(
            method="POST", 
            url="/v1/consensus/reconcile", 
            data=data, 
            idempotency_key=idempotency_key
        )

class Consensus(SyncAPIResource, BaseConsensusMixin):
    """Consensus API wrapper for synchronous operations"""

    def __init__(self, client: Any) -> None:
        super().__init__(client=client)
        self.completions = Completions(client=client)
        self.responses = Responses(client=client)

    def reconcile(
        self,
        list_dicts: List[Dict[str, Any]],
        reference_schema: Optional[Dict[str, Any]] = None,
        mode: Literal["direct", "aligned"] = "direct",
        idempotency_key: str | None = None,
    ) -> ReconciliationResponse:
        """
        Reconcile multiple dictionaries to produce a single unified consensus dictionary.
        
        Args:
            list_dicts: List of dictionaries to reconcile
            reference_schema: Optional schema to validate dictionaries against
            mode: Mode for consensus computation ("direct" or "aligned")
            idempotency_key: Optional idempotency key for the request
            
        Returns:
            Dict containing the consensus dictionary and consensus likelihoods
            
        Raises:
            UiformAPIError: If the API request fails
        """
        request = self._prepare_reconcile(
            list_dicts, 
            reference_schema, 
            mode, 
            idempotency_key
        )
        response = self._client._prepared_request(request)
        return ReconciliationResponse.model_validate(response)


class AsyncConsensus(AsyncAPIResource, BaseConsensusMixin):
    """Consensus API wrapper for asynchronous operations"""

    def __init__(self, client: Any) -> None:
        super().__init__(client=client)
        self.completions = AsyncCompletions(client=client)
        self.responses = AsyncResponses(client=client)

    
    async def reconcile(
        self,
        list_dicts: List[Dict[str, Any]],
        reference_schema: Optional[Dict[str, Any]] = None,
        mode: Literal["direct", "aligned"] = "direct",
        idempotency_key: str | None = None,
    ) -> ReconciliationResponse:
        """
        Reconcile multiple dictionaries to produce a single unified consensus dictionary asynchronously.
        
        Args:
            list_dicts: List of dictionaries to reconcile
            reference_schema: Optional schema to validate dictionaries against
            mode: Mode for consensus computation ("direct" or "aligned")
            idempotency_key: Optional idempotency key for the request
            
        Returns:
            Dict containing the consensus dictionary and consensus likelihoods
            
        Raises:
            UiformAPIError: If the API request fails
        """
        request = self._prepare_reconcile(
            list_dicts, 
            reference_schema, 
            mode, 
            idempotency_key
        )
        response = await self._client._prepared_request(request)

        return ReconciliationResponse.model_validate(response)



from pydantic import BaseModel, Field


class ReconciliationResponse(BaseModel):
    consensus_dict: dict = Field(
        description="The consensus dictionary containing the reconciled values from the input dictionaries."
    )
    likelihoods: dict = Field(
        description="A dictionary containing the likelihood/confidence scores for each field in the consensus dictionary."
    )



from typing import get_args

from ..types.ai_models import AIProvider, GeminiModel, OpenAIModel, xAI_Model


def find_provider_from_model(model: str) -> AIProvider:
    if model in get_args(OpenAIModel):
        return "OpenAI"
    elif ":" in model:
        # Handle fine-tuned models
        ft, base_model, model_id = model.split(":", 2)
        if base_model in get_args(OpenAIModel):
            return "OpenAI"
    # elif model in get_args(AnthropicModel):
    #    return "Anthropic"
    elif model in get_args(xAI_Model):
        return "xAI"
    elif model in get_args(GeminiModel):
        return "Gemini"
    raise ValueError(f"Could not determine AI provider for model: {model}")


def assert_valid_model_extraction(model: str) -> None:
    if model in get_args(OpenAIModel):
        return
    elif ":" in model:
        # Handle fine-tuned models
        ft, base_model, model_id = model.split(":", 2)
        if base_model in get_args(OpenAIModel):
            return
    # elif model in get_args(AnthropicModel):
    #    return
    elif model in get_args(xAI_Model):
        return
    elif model in get_args(GeminiModel):
        return
    raise ValueError(
        f"Invalid model for extraction: {model}.\nValid OpenAI models: {get_args(OpenAIModel)}\n"
        # f"Valid Anthropic models: {get_args(AnthropicModel)}\n"
        # f"Valid xAI models: {get_args(xAI_Model)}\n"
        # f"Valid Gemini models: {get_args(GeminiModel)}"
    )


def assert_valid_model_batch_processing(model: str) -> None:
    """Assert that the model is either a standard OpenAI model or a valid fine-tuned model.

    Valid formats:
    - Standard model: Must be in OpenAIModel
    - Fine-tuned model: Must be {base_model}:{id} where base_model is in OpenAIModel

    Raises:
        ValueError: If the model format is invalid
    """
    if model in get_args(OpenAIModel):
        return

    try:
        ft, base_model, model_id = model.split(":", 2)
        if base_model not in get_args(OpenAIModel):
            raise ValueError(f"Invalid base model in fine-tuned model '{model}'. Base model must be one of: {get_args(OpenAIModel)}")
        if not model_id or not model_id.strip():
            raise ValueError(f"Model ID cannot be empty in fine-tuned model '{model}'")
    except ValueError as e:
        if ":" not in model:
            raise ValueError(
                f"Invalid model format: {model}. Must be either:\n"
                f"1. A standard model: {get_args(OpenAIModel)}\n"
                f"2. A fine-tuned model in format 'base_model:id' where base_model is one of the standard models"
            ) from None
        raise


def assert_valid_model_schema_generation(model: str) -> None:
    """Assert that the model is either a standard OpenAI model or a valid fine-tuned model.

    Valid formats:
    - Standard model: Must be in OpenAIModel
    - Fine-tuned model: Must be {base_model}:{id} where base_model is in OpenAIModel

    Raises:
        ValueError: If the model format is invalid
    """
    if model in get_args(OpenAIModel):
        return

    try:
        ft, base_model, model_id = model.split(":", 2)
        if base_model not in get_args(OpenAIModel):
            raise ValueError(f"Invalid base model in fine-tuned model '{model}'. Base model must be one of: {get_args(OpenAIModel)}")
        if not model_id or not model_id.strip():
            raise ValueError(f"Model ID cannot be empty in fine-tuned model '{model}'")
    except ValueError as e:
        if ":" not in model:
            raise ValueError(
                f"Invalid model format: {model}. Must be either:\n"
                f"1. A standard model: {get_args(OpenAIModel)}\n"
                f"2. A fine-tuned model in format 'base_model:id' where base_model is one of the standard models"
            ) from None
        raise



from typing import Any, get_args

from openai import AsyncOpenAI, OpenAI

from .._resource import AsyncAPIResource, SyncAPIResource
from ..types.ai_models import OpenAIModel


class FineTuningJobs(SyncAPIResource):
    def create(self, training_file: str, model: str) -> Any:
        """Create a new fine-tuning job"""
        openai_client = OpenAI(api_key=self._client.headers["OpenAI-Api-Key"])
        assert model in get_args(OpenAIModel), f"Model {model} is not supported. Supported models are: {get_args(OpenAIModel)}"
        return openai_client.fine_tuning.jobs.create(training_file=training_file, model=model)

    def retrieve(self, fine_tuning_job_id: str) -> Any:
        """Retrieve status of a fine-tuning job"""
        openai_client = OpenAI(api_key=self._client.headers["OpenAI-Api-Key"])
        return openai_client.fine_tuning.jobs.retrieve(fine_tuning_job_id)


class AsyncFineTuningJobs(AsyncAPIResource):
    async def create(self, training_file: str, model: str) -> Any:
        """Create a new fine-tuning job"""
        assert model in get_args(OpenAIModel), f"Model {model} is not supported. Supported models are: {get_args(OpenAIModel)}"
        async with AsyncOpenAI(api_key=self._client.headers["OpenAI-Api-Key"]) as openai_client:
            return await openai_client.fine_tuning.jobs.create(training_file=training_file, model=model)

    async def retrieve(self, fine_tuning_job_id: str) -> Any:
        """Retrieve status of a fine-tuning job"""
        async with AsyncOpenAI(api_key=self._client.headers["OpenAI-Api-Key"]) as openai_client:
            return await openai_client.fine_tuning.jobs.retrieve(fine_tuning_job_id)


class FineTuning(SyncAPIResource):
    """Fine-tuning jobs API wrapper"""

    _jobs: FineTuningJobs

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self._jobs = FineTuningJobs(self._client)  # Initialize the Jobs instance with the client

    @property
    def jobs(self) -> FineTuningJobs:
        """Expose your private _jobs attribute through a property."""
        return self._jobs


class AsyncFineTuning(AsyncAPIResource):
    """Fine-tuning jobs Asyncronous API wrapper"""

    _jobs: AsyncFineTuningJobs

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self._jobs = AsyncFineTuningJobs(self._client)  # Initialize the Jobs instance with the client

    @property
    def jobs(self) -> AsyncFineTuningJobs:
        """Expose your private _jobs attribute through a property."""
        return self._jobs



# from typing import Any
# from pathlib import Path
# import json

# from .._resource import SyncAPIResource, AsyncAPIResource
# from .._utils.json_schema import load_json_schema
# from ..types.jobs import JobResponse
# from ..types.jobs.prompt_optimization import PromptOptimizationObject, PromptOptimizationProps, PromptOptimizationJobInputData, PromptOptimizationJob

# MAX_TRAINING_SAMPLES = 10

# class PromptOptimizationJobsMixin:
#     def prepare_create(self, json_schema: dict[str, Any] | Path | str,
#                        training_file: str,
#                        schema_optimization_props: dict[str, Any]) -> PromptOptimizationJob:

#         optimization_objects = []
#         with open(training_file, "r") as f:
#             for line in f:
#                 optimization_objects.append(PromptOptimizationObject(**json.loads(line)))


#         job = PromptOptimizationJob(
#             job_type="prompt-optimization",
#             input_data=PromptOptimizationJobInputData(
#                 json_schema=load_json_schema(json_schema),
#                 optimization_objects=optimization_objects[:MAX_TRAINING_SAMPLES],
#                 schema_optimization_props=PromptOptimizationProps.model_validate(schema_optimization_props)
#             )
#         )
#         return job

# class PromptOptimizationJobs(SyncAPIResource, PromptOptimizationJobsMixin):
#     def create(self, json_schema: dict[str, Any] | Path | str, training_file: str, schema_optimization_props: dict[str, Any]) -> JobResponse:
#         """Create a new prompt optimization job"""

#         request_data = self.prepare_create(json_schema, training_file, schema_optimization_props)
#         response = self._client._request("POST", "/v1/jobs", data=request_data.model_dump(mode="json"))
#         return JobResponse.model_validate(response)


#     def retrieve(self, job_id: str) -> Any:
#         """Retrieve status of a prompt optimization job"""
#         response = self._client._request("GET", f"/v1/jobs/{job_id}")
#         return JobResponse.model_validate(response)

# class AsyncPromptOptimizationJobs(AsyncAPIResource, PromptOptimizationJobsMixin):
#     async def create(self, json_schema: dict[str, Any] | Path | str, training_file: str, schema_optimization_props: dict[str, Any]) -> Any:
#         """Create a new prompt optimization job"""

#         request_data = self.prepare_create(json_schema, training_file, schema_optimization_props)
#         response = await self._client._request("POST", "/v1/jobs/", data=request_data.model_dump(mode="json"))
#         return JobResponse.model_validate(response)

#     async def retrieve(self, job_id: str) -> Any:
#         """Retrieve status of a prompt optimization job"""

#         response = await self._client._request("GET", f"/v1/jobs/{job_id}")
#         return JobResponse.model_validate(response)


# class PromptOptimization(SyncAPIResource):
#     """Prompt optimization jobs API wrapper"""
#     _jobs: PromptOptimizationJobs

#     def __init__(self, *args: Any, **kwargs: Any):
#         super().__init__(*args, **kwargs)
#         self._jobs = PromptOptimizationJobs(client=self._client)

# class AsyncPromptOptimization(AsyncAPIResource):
#     """Prompt optimization jobs Asyncronous API wrapper"""
#     _jobs: AsyncPromptOptimizationJobs

#     def __init__(self, *args: Any, **kwargs: Any):
#         super().__init__(*args, **kwargs)
#         self._jobs = AsyncPromptOptimizationJobs(client=self._client)



import datetime
from typing import Any, Dict, Literal

import nanoid  # type: ignore
from pydantic import BaseModel, Field

from ..modalities import Modality


class AnnotationParameters(BaseModel):
    model: str
    modality: Modality | None = "native"
    image_resolution_dpi: int = 96
    browser_canvas: Literal['A3', 'A4', 'A5'] = 'A4'
    temperature: float = 0.0


class Annotation(BaseModel):
    file_id: str = Field(description="ID of the file that the annotation belongs to")
    parameters: AnnotationParameters = Field(description="Parameters used for the annotation")
    data: Dict[str, Any] = Field(default_factory=dict, description="Data of the annotation")
    schema_id: str = Field(description="ID of the schema used for the annotation")
    organization_id: str = Field(description="ID of the organization that owns the annotation")
    updated_at: datetime.datetime = Field(default_factory=lambda: datetime.datetime.now(datetime.timezone.utc), description="Timestamp for when the annotation was last updated")



from typing import Any

from openai.types.chat.chat_completion_reasoning_effort import ChatCompletionReasoningEffort
from pydantic import BaseModel, ConfigDict, Field

from .._utils.ai_models import find_provider_from_model
from .ai_models import AIProvider
from .chat import ChatCompletionUiformMessage


from openai.types.shared_params.response_format_json_schema import ResponseFormatJSONSchema

class UiChatCompletionsRequest(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    model: str = Field(..., description="Model used for chat completion")
    messages: list[ChatCompletionUiformMessage] = Field(..., description="Messages to be parsed")
    response_format: ResponseFormatJSONSchema = Field(..., description="response format used to validate the output data.")
    temperature: float = Field(default=0.0, description="Temperature for sampling. If not provided, the default temperature for the model will be used.", examples=[0.0])
    reasoning_effort: ChatCompletionReasoningEffort = Field(
        default="medium", description="The effort level for the model to reason about the input data. If not provided, the default reasoning effort for the model will be used."
    )
    # Regular fields
    stream: bool = Field(default=False, description="If true, the extraction will be streamed to the user using the active WebSocket connection")
    seed: int | None = Field(default=None, description="Seed for the random number generator. If not provided, a random seed will be generated.", examples=[None])
    n_consensus: int = Field(default=1, description="Number of consensus models to use for extraction. If greater than 1 the temperature cannot be 0.")

    @property
    def provider(self) -> AIProvider:
        """
        Determines the AI provider based on the model specified.

        Returns:
            AIProvider: The AI provider corresponding to the given model.
        """
        return find_provider_from_model(self.model)




class UiChatCompletionsParseRequest(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    model: str = Field(..., description="Model used for chat completion")
    messages: list[ChatCompletionUiformMessage] = Field(..., description="Messages to be parsed")
    json_schema: dict[str, Any] = Field(..., description="JSON schema format used to validate the output data.")
    temperature: float = Field(default=0.0, description="Temperature for sampling. If not provided, the default temperature for the model will be used.", examples=[0.0])
    reasoning_effort: ChatCompletionReasoningEffort = Field(
        default="medium", description="The effort level for the model to reason about the input data. If not provided, the default reasoning effort for the model will be used."
    )
    # Regular fields
    stream: bool = Field(default=False, description="If true, the extraction will be streamed to the user using the active WebSocket connection")
    seed: int | None = Field(default=None, description="Seed for the random number generator. If not provided, a random seed will be generated.", examples=[None])
    n_consensus: int = Field(default=1, description="Number of consensus models to use for extraction. If greater than 1 the temperature cannot be 0.")

    @property
    def provider(self) -> AIProvider:
        """
        Determines the AI provider based on the model specified.

        Returns:
            AIProvider: The AI provider corresponding to the given model.
        """
        return find_provider_from_model(self.model)

from typing import Optional, Union
from openai.types.shared_params.reasoning import Reasoning
from openai.types.responses.response_input_param import ResponseInputParam
from openai.types.responses.response_text_config_param import ResponseTextConfigParam

class UiChatResponseCreateRequest(BaseModel):
    input: Union[str, ResponseInputParam] = Field(..., description="Input to be parsed")
    instructions: Optional[str] = None

    model_config = ConfigDict(arbitrary_types_allowed=True)
    model: str = Field(..., description="Model used for chat completion")
    temperature: Optional[float] = Field(default=0.0, description="Temperature for sampling. If not provided, the default temperature for the model will be used.", examples=[0.0])
    reasoning: Optional[Reasoning] = Field(default=None, description="The effort level for the model to reason about the input data. If not provided, the default reasoning effort for the model will be used.")
    
    stream: Optional[bool] = Field(default=False, description="If true, the extraction will be streamed to the user using the active WebSocket connection")
    seed: int | None = Field(default=None, description="Seed for the random number generator. If not provided, a random seed will be generated.", examples=[None])
    text: ResponseTextConfigParam = Field(default={"format": {"type": "text"}}, description="Format of the response")

    n_consensus: int = Field(default=1, description="Number of consensus models to use for extraction. If greater than 1 the temperature cannot be 0.")

    @property
    def provider(self) -> AIProvider:
        """
        Determines the AI provider based on the model specified.

        Returns:
            AIProvider: The AI provider corresponding to the given model.
        """
        return find_provider_from_model(self.model)




from typing import Any, Optional

from pydantic import BaseModel, EmailStr

from uiform.types.documents.extractions import UiParsedChatCompletion

from ..mime import BaseMIMEData, MIMEData


class WebhookRequest(BaseModel):
    completion: UiParsedChatCompletion
    user: Optional[EmailStr] = None
    file_payload: MIMEData
    metadata: Optional[dict[str, Any]] = None


class BaseWebhookRequest(BaseModel):
    completion: UiParsedChatCompletion
    user: Optional[EmailStr] = None
    file_payload: BaseMIMEData
    metadata: Optional[dict[str, Any]] = None
